/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: simulation_ctmc.c,v $
*	$Revision: 1.87 $
*	$Date: 2008/06/02 16:28:43 $
*	$Author: i.zapreev $
*
*	MRMC is a model checker for discrete-time and continuous-time Markov
*	reward models. It supports reward extensions of PCTL and CSL (PRCTL 
*	and CSRL), and allows for the automated verification of properties
*	concerning long-run and instantaneous rewards as well as cumulative
*	rewards.
*	
* 	Copyright (C) The University of Twente, 2004-2008.
* 	Copyright (C) RWTH Aachen, 2008-2009.
*	Authors: Ivan Zapreev, Christina Jansen
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
*	MA  02110-1301, USA.
*
*	Main contact:
*		Lehrstuhl für Informatik 2, RWTH Aachen University
*		Ahornstrasse 55, 52074 Aachen, Germany
*		E-mail: info@mrmc-tool.org
*	
*	Old ontact:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is a header file for the simulation engine
*	here we intend to define the CTMC-related functions and data 
*	structures.
*/

#include "simulation_ctmc.h"

#include "simulation_common.h"
#include "simulation_utils.h"
#include "transient_common.h"
#include "bscc.h"
#include "gsl/gsl_cdf.h"

/*This macro computes the N'th root of X */
#define ROOT(X, N) (pow(X, (double)1.0/N))

/**********************************************************************/
/********THE FUNCTIONS FOR SIMULATING UNBOUNDED UNTIL CTMC*************/
/**********************************************************************/

/**
* This method updates the confidence interval borders for states with 
* trivially computed probability 1.0. It also allocates the conf.
* int. arrays, in case it was not done beforehand, i.e. the complete
* initial state set consists only of states, whose probabilities can
* be computed trivially.
* NOTE: pAlwaysBitSet is allowed to be NULL!
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param ppProbCILeftBorder for storing the left conf. int. border
* @param ppProbCIRightBorder for storing the right conf. int. border
* @param pAlwaysBitSet the states that satisfy the formula with probability >= 1.0
* @param initial_state the initial state index if isSimOneInitState == TRUE
*/
static inline void updateConfIntAlwaysStates(const BOOL isSimOneInitState, double ** ppProbCILeftBorder,
		double ** ppProbCIRightBorder, int * pResultSize, const bitset * pAlwaysBitSet, const int initial_state)
{
	int curr_index, state_space_size;
	if( isSimOneInitState ){
		/* Is the conf. int. array allocated yet? */
		if( ( *ppProbCILeftBorder ) == NULL ){
			*pResultSize = 1;
			*ppProbCILeftBorder = (double * ) calloc( 1, sizeof(double) );
			*ppProbCIRightBorder = (double * ) calloc( 1, sizeof(double) );
		}
		/* If we simulate just one initial state, test if it belongs to pAlwaysBitSet */
		if( ( pAlwaysBitSet != NULL ) && get_bit_val( pAlwaysBitSet, initial_state ) ){
			(*ppProbCILeftBorder)[0] = (*ppProbCIRightBorder)[0] = 1.0;
		} 
	} else {
		/* Is the conf. int. array allocated yet? */
		if( ( *ppProbCILeftBorder ) == NULL ){
			state_space_size = pAlwaysBitSet->n;
			*pResultSize = state_space_size;
			*ppProbCILeftBorder = (double * ) calloc( state_space_size, sizeof(double) );
			*ppProbCIRightBorder = (double * ) calloc( state_space_size, sizeof(double) );
		}		
		/* All pAlwaysBitSet states satisfy the formula with probability >= 1.0 */
		if( pAlwaysBitSet != NULL ){
			/* Set the conf. int. borders to 1.0 for every state, which trivially */
			/* fulfills the formula. */
			curr_index = -1;
			while( ( curr_index =  get_idx_next_non_zero( pAlwaysBitSet, curr_index ) ) != -1 ){
				(*ppProbCILeftBorder)[curr_index]  = (*ppProbCIRightBorder)[curr_index] = 1.0;
			}
		}
	}
}

/**
* This function is an implementation of the unboundedUntil function from the PhD Thesis
* of Ivan S. Zapreev. The function does the model checking of unbounded until operator
* on the provided CTMC and uses the runtime-simulation settings from the "simulation.h"
* @param pStateSpace the sparse matrix of the embedded DTMC with good and bad
*			states made absorbing
* @param confidence this is is confidence we should use for simulations.
*			Note that since the formulas can be nested one can not use the
*			overall confidence set in simulation.h
* @param pPhiBitSet the bitsets containing all the good absorbing states
* @param pPsiBitSet the bitsets containing all the transient states
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param initial_state the initial state index if isSimOneInitState == TRUE
*
* NOTE: Below we have pointers to the pointers, these are going to be the return values
*
* @param ppYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param ppNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param ppCiLeftBorders will contain the left conf. int. borders
* @param ppCiRightBorders will contain the right conf. int. borders
* @param pResultSize the pointer to the size of arrays *ppCiLeftBorders and *ppCiRightBorders
* @param pMaxNumUsedObserv the return-value variable that will store the maximum number
* (over all initial states) of used observations
*/
extern void modelCheckUnboundedUntilCTMC( const sparse * pStateSpace, const double * pCTMCRowSums,
						const double confidence, const bitset *pPhiBitSet,
						const bitset * pPsiBitSet, bitset ** ppYesBitsetResult,
						bitset ** ppNoBitsetResult, double ** ppProbCILeftBorder,
						double ** ppProbCIRightBorder, int * pResultSize,
						const int comparator, const double prob_bound,
						const int initial_state, const BOOL isSimOneInitState,
						unsigned int * pMaxNumUsedObserv ){
	int *pTransientStateIdx=NULL;
	/* This vector will store the exit rates for the elements of pTransientBitSet*/
	double * pExitRatesOfTransStates = NULL;
	
	/* 0: Sort out the 1.0 and 0.0 prob reachabile states */
	bitset * pEUBitSet = get_exist_until( pStateSpace, pPhiBitSet, pPsiBitSet );
	bitset * pAUBitSet = get_always_until( pStateSpace, pPhiBitSet, pPsiBitSet, pEUBitSet );
	
	/* NOTE: Using pEUBitSet there is NO NEED to compute BSCCs! */
	/* Because "pEUBitSet = get_exist_until( pStateSpace, pPhiBitSet, pPsiBitSet )" */
	/* returns the states from which it is possible to go into Psi states via Phi states */
	
	/* 1: Compute the states that has to be absorbing */
	/* 1.1: At this point pBadBitSet contains states from which */
	/* psi states are not reachable via phi states */
	bitset *pBadBitSet = not( pEUBitSet );
	/* 1.2: At this point pTransientBitSet contains all absorbing states, i.e. */
	/* Good and Bad absorbing states (including Phi BSCC states) */
	bitset * pTransientBitSet = or( pAUBitSet, pBadBitSet );
	/* 1.3: At this point pTransientBitSet contains only allowed transient states */
	/* from which there is a way to go to Psi states via Phi states with prob < 1.0 */
	not_result( pTransientBitSet );
	
	/* print_mtx_sparse(pStateSpace); */
	
	/* 3: Make states absorbing */
	pTransientStateIdx = count_set( pTransientBitSet );
	
	/* 4: Obtain the embedded DTMC (WITHOUT THE SELF LOOPS!!!) */
	pExitRatesOfTransStates = make_embedded_dtmc_out_of_rate_mtx_vs( pStateSpace, pCTMCRowSums, pTransientStateIdx );
	
	/* print_mtx_sparse(pStateSpace); */
	
	/* 5: Construct the pGoodStates and pTransientStates sets */
	/* NOTE: They are already constructed, namely: pAUBitSet and pTransientBitSet */
	
	/* 6: Do simulations */
	modelCheckStatesCommon( pStateSpace, confidence, ppYesBitsetResult, ppNoBitsetResult, ppProbCILeftBorder,
				ppProbCIRightBorder, pResultSize, comparator, prob_bound, initial_state,
				isSimOneInitState, pTransientBitSet, pMaxNumUsedObserv, modelCheckOneStateUUCommon,
				pTransientBitSet, pAUBitSet );
	
	/* 7.1: Update ppYesBitsetResult and ppNoBitsetResult with pAUBitSet and pBadBitSet states. */
	considerAlwaysAndNeverStates( isSimOneInitState, *ppYesBitsetResult, *ppNoBitsetResult,
					pAUBitSet, pBadBitSet, initial_state, comparator, prob_bound );
	/* 7.2: Update ppProbCILeftBorder and ppProbCIRightBorder with probability 1.0 for pAUBitSet states. */
	updateConfIntAlwaysStates( isSimOneInitState, ppProbCILeftBorder, ppProbCIRightBorder, pResultSize,
					 pAUBitSet, initial_state );
	
	/* 8: Restore the state space */
	/* NOTE: not required for diagonal elements */
	restore_rate_mtx_out_of_embedded_dtmc_vs( pStateSpace, pTransientStateIdx, pExitRatesOfTransStates );
	
	/* print_mtx_sparse(pStateSpace); */
	
	/* 9: Free the resources */
	free( pExitRatesOfTransStates );
	free( pTransientStateIdx );
	free_bitset( pEUBitSet );
	free_bitset( pAUBitSet );
	free_bitset( pBadBitSet );
	free_bitset( pTransientBitSet );
}

/*********************************************************************/
/********THE FUNCTIONS FOR SIMULATING INTERVAL UNTIL CTMC*************/
/*********************************************************************/

/**
 * This function computes when we will leave the state.
 * In addition to that for an absorbing current state the time of leaving the state is 
 * set to be right_time_bound + 1.0 in order to avoid further simulations
 * @param current_exit_rate the exit rate of the current state 
 * @param right_time_bound the right time bound of the until
 * @return the next exit time
 */
static inline double computeExitTime( const double current_exit_rate, const double right_time_bound ){
	/* If we are not in an absorbing state */
	if( current_exit_rate != 0.0 ){
		/* Compute when we leave the current state */
		return generateRandNumberExp( current_exit_rate );
	} else {
		/* If we are absorbed in a Phi state which might be also a Psi state */
		/* then we just go beyond the time interval to stop further iterations */
		return right_time_bound + 1.0;
	}
}

/**
 * This function computes to which state we will go.
 * @param sOutgoingTrans the row of the sparse matrix
 * @param ncols the number of transitions for the current state
 * WARNING: The self loops are not taken into account, we expect there to be NONE
 * @param current_obs_state the pointer to the next state index 
 * @return the next state
 */
static inline int computeNextState( const values sOutgoingTrans, const int ncols, const int current_obs_state ){
	/* If we are not in an absorbing state */
	if( ncols != 0 ){
		/* Compute to what state we will go */
		return generateRandNumberDiscrete( sOutgoingTrans.col, sOutgoingTrans.val, ncols );
	} else {
		return current_obs_state;
	}
}

/**
* This function is designed to simulate the new sample observations for interval
* until CTMC: "Phi U[tl, tr] Psi".
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC
* @param pSampleVecIntUntil the sample vector obtained on the previous iteration
* @param old_sample_size the old sample size
* @param pNotPhiAndNotPsiBitSet the bitsets containing all
*	not Phi and not Psi states
* @param pPhiBitSet the bitsets containing all the Phi
* @param pPsiBitSet the bitsets containing all the Psi
* @param left_time_bound the left time bound "tl"
* @param right_time_bound the right time bound "tr"
* @param pExitRatesOfAllowedStates the array of exit rates for the "not pNotPhiAndNotPsiBitSet"
*	states, not that its size corresponds to the dimensions of pStateSpace
*/
static void simulateIntUntilSampleCTMC( const sparse * pStateSpace, PTSampleVecIntUntil pSampleVecIntUntil,
					const int old_sample_size, const bitset * pNotPhiAndNotPsiBitSet,
					const bitset * pPhiBitSet, const bitset * pPsiBitSet,
					const double left_time_bound, const double right_time_bound,
					const double * pExitRatesOfAllowedStates ){
	int i, current_obs_state;
	double current_obs_enter_time, current_obs_exit_time;
	/* Will store the number of newly simulated observations */
	unsigned int newlyVisitedStates = 0;
	
	/* Cast to the parent structure which contains initial_state and curr_sample_size */
	PTSampleVec pSampleVecIntUntilBase = (PTSampleVec) pSampleVecIntUntil;
	
	/* For all the newly added observation in the sample */
	for( i = 0; i < ( pSampleVecIntUntilBase->curr_sample_size - old_sample_size ) ; i++ ){
		/* Take the initial state index */
		current_obs_state = pSampleVecIntUntilBase->initial_state;
		/* We start in the initial state at time 0.0 */
		current_obs_enter_time = 0.0;
		/* Compute when we leave the initial state */
		current_obs_exit_time = computeExitTime( pExitRatesOfAllowedStates[current_obs_state], right_time_bound );
		
		/* printf("\n----------------------------------------------------\n"); */
		/* printf("I: (%s) state %d exit at %e \n", (isInitialPartFailed? "-": "+"),
				current_obs_state+1, current_obs_exit_time ); */
		
		/* Iterate while we go through Phi states until left_time_bound */
		while( ( current_obs_exit_time <= left_time_bound ) && get_bit_val( pPhiBitSet, current_obs_state ) ){
			/* Go through the states until */
			current_obs_state = computeNextState( pStateSpace->val[ current_obs_state ],
						pStateSpace->ncols[ current_obs_state ], current_obs_state );
			current_obs_enter_time = current_obs_exit_time;
			current_obs_exit_time += computeExitTime( pExitRatesOfAllowedStates[current_obs_state],
									right_time_bound );
			/* The next state was simulated */
			newlyVisitedStates++;
		}
		
		/* printf("M: (%s) state %d exit at %e \n", (isInitialPartFailed? "-": "+"),
				current_obs_state+1, current_obs_exit_time ); */
		
		if( ( current_obs_enter_time == left_time_bound ) || get_bit_val( pPhiBitSet, current_obs_state ) ){
			/* Iterate while we go through Phi states */
			while( ! get_bit_val( pNotPhiAndNotPsiBitSet, current_obs_state ) &&
				! get_bit_val( pPsiBitSet, current_obs_state) &&
				( current_obs_exit_time <= right_time_bound ) ){
				/* While we are not in a bad state not in the target state and before right_time_bound */
				current_obs_state = computeNextState( pStateSpace->val[ current_obs_state ],
									pStateSpace->ncols[ current_obs_state ],
									current_obs_state );
				current_obs_exit_time += computeExitTime( pExitRatesOfAllowedStates[current_obs_state],
									right_time_bound );
				/* The next state was simulated */
				newlyVisitedStates++;
			}
			
			/* Update the sum_good field if we ended up in a Psi state in the right time */
			if( get_bit_val( pPsiBitSet, current_obs_state) ){
				pSampleVecIntUntil->sum_good += 1;
			}
			
			/* printf("F: (%s) state %d exit at %e \n",
				( get_bit_val( pPsiBitSet, current_obs_state) ? "+": "-"),
				current_obs_state+1, current_obs_exit_time ); */
		}
	}
	
	/* printSampleVectorIntUntil( pSampleVecIntUntil ); */
	
	/* Update the number of states visited while preparing this sample */
	pSampleVecIntUntilBase->num_visited_states += newlyVisitedStates;
}

/**
* This function is designed to extended and simulate the sample for interval
* until CTMC: "Phi U[tl, tr] Psi". This function is an implementation of
* the "extendSample" function from the PhD Thesis of Ivan S. Zapreev. It adds
* new observations to the sample and simulates.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC
* @param pSampleVecIntUntil the sample vector obtained on the previous iteration
* @param new_sample_size the new sample size
* @param pNotPhiAndNotPsiBitSet the bitsets containing all 
*	not Phi and not Psi states
* @param pPhiBitSet the bitsets containing all the Phi
* @param pPsiBitSet the bitsets containing all the Psi
* @param left_time_bound the left time bound "tl"
* @param right_time_bound the right time bound "tr"
* @param pExitRatesOfAllowedStates the array of exit rates for the "not pNotPhiAndNotPsiBitSet"
*	states, not that its size corresponds to the dimensions of pStateSpace
*/
static void simulateSampleVectorIntUntilCTMC( const sparse * pStateSpace, PTSampleVecIntUntil pSampleVecIntUntil,
					const int new_sample_size, const bitset * pNotPhiAndNotPsiBitSet,
					const bitset * pPhiBitSet, const bitset * pPsiBitSet, const double left_time_bound,
					const double right_time_bound, const double * pExitRatesOfAllowedStates ){
	IF_SAFETY( pSampleVecIntUntil != NULL )
		/* Cast to the parent structure which contains initial_state and curr_sample_size */
		PTSampleVec pSampleVecIntUntilBase = (PTSampleVec) pSampleVecIntUntil;
		
		IF_SAFETY( ( pNotPhiAndNotPsiBitSet != NULL ) && ( pPhiBitSet != NULL ) && ( pPsiBitSet != NULL ) )
			IF_SAFETY( pStateSpace != NULL )
				IF_SAFETY( pExitRatesOfAllowedStates != NULL )
					IF_SAFETY( pSampleVecIntUntilBase->curr_sample_size < new_sample_size )
						const int old_sample_size = pSampleVecIntUntilBase->curr_sample_size;
						
						/* Extend the sample size */
						extendSampleVectorIntUntil( pSampleVecIntUntil, new_sample_size );
						
						/*Do simulations for the newly added observations */
						simulateIntUntilSampleCTMC( pStateSpace, pSampleVecIntUntil, old_sample_size,
									pNotPhiAndNotPsiBitSet, pPhiBitSet, pPsiBitSet,
									left_time_bound, right_time_bound, pExitRatesOfAllowedStates );
					ELSE_SAFETY
						printf("ERROR: The sample size is not going to increase, there is nothing to simulate.\n");
						exit(1);
					ENDIF_SAFETY
				ELSE_SAFETY
					printf("ERROR: The exit-rates array is NULL.\n");
					exit(1);
				ENDIF_SAFETY
			ELSE_SAFETY
				printf("ERROR: The sparse matrix is NULL.\n");
				exit(1);
			ENDIF_SAFETY
		ELSE_SAFETY
			printf("ERROR: The Phi, Psi or not Phi and not Psi states are undefined.\n");
			exit(1);
		ENDIF_SAFETY
	ELSE_SAFETY
		printf("ERROR: The sample that has to be simulated further is NULL.\n");
		exit(1);
	ENDIF_SAFETY
}

/**
* Model checks the interval-until operator "Phi U[tl, tr] Psi" for one initial state only.
* Note that this is a univesral procedure that assumes that we work with the embedded DTMC.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC with good and bad
*			states made absorbing
* @param initial_state the initial state for simulations
* @param indiff_width the width of the indifference region
* @param confidence the desired confidence of the answer
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param arr_index indicates the index in arrays pCiLeftBorders and pCiRightBorders
*			that corresponds to the initial_state
*
* NOTE: The parameters below will contain the return values
*
* @param pCiLeftBorders for storing the left conf. int. border
* @param pCiRightBorders for storing the right conf. int. border
* @param pYesBitSet contains curr_state if it satisfies the formula
* @param pNoBitSet contains curr_state if it does not satisfy the formula
* @param pNumUsedObserv the pointer to the return-value variable that will store the
*		number of observations used for simulating this one state.
* @param pArguments contains the following (ordered!):
*		1. pNotPhiAndNotPsiBitSet the bitsets containing all 
*			not Phi and not Psi states
* 		2. pPhiBitSet the bitsets containing all the Phi
*		3. pPsiBitSet the bitsets containing all the Psi
*		4. left_time_bound the left time bound "tl"
*		5. right_time_bound the right time bound "tr"
*		6. pExitRatesOfAllowedStates the array of exit rates for the
*			"not pNotPhiAndNotPsiBitSet" states, not that its size corresponds
*			to the dimensions of pStateSpace
*/
static void modelCheckOneStateIUCTMC( const sparse* pStateSpace, const int initial_state, const double indiff_width,
					const double confidence, const int comparator, const double prob_bound,
					const int arr_index, double *pCiLeftBorders, double *pCiRightBorders,
					bitset * pYesBitSet, bitset * pNoBitSet, unsigned int *pNumUsedObserv,
					va_list pArguments ){
	/* Contain the pNotPhiAndNotPsiBitSet, pPhiBitSet and pPsiBitSet states respectively */
	const bitset * pNotPhiAndNotPsiBitSet = va_arg( pArguments, bitset * );
	const bitset * pPhiBitSet = va_arg( pArguments, bitset * );
	const bitset * pPsiBitSet = va_arg( pArguments, bitset * );
	/* Contain the left and the right time bounds respectively */
	const double left_time_bound = va_arg( pArguments, double );
	const double right_time_bound = va_arg( pArguments, double );
	/* Contains the exit rates for the "not pNotPhiAndNotPsiBitSet" states */
	const double * pExitRatesOfAllowedStates = va_arg( pArguments, double * );
	
	/* The model checking result*/
	TV_LOGIC mc_result = TVL_NN;
	/* The conf. int. left and right borders */
	double leftBorder = 0.0, rightBorder = 1.0;
	
	/* Compute the general confidence zeta value */
	/* We compute it for the 1 - beta/2 with confidence = 1 - beta */
	const double gen_conf_zeta = gsl_cdf_ugaussian_Pinv( 0.5 * ( 1 + confidence ) );
	/* NOTE: The confidence we have is used to derive z for the left and the */
	/* right border BUT the TRUE/FALSE answer is based only on one border of */
	/* the conf. int., i.e. we could use gsl_cdf_ugaussian_Pinv( confidence ) */
	/* for model checking, but then again we have a different confidence for the */
	/* produced conf. int. this is bad because then the user does not know that. */
	
	/* Initialize the sample size parameters */
	int sample_size = getSimMinSampleSize();
	const int sample_size_step = getSimSampleSizeStep();
	const int max_sample_size = getSimMaxSampleSize();
	
	/* Allocate an empty sample vector */
	PTSampleVecIntUntil pSampleVecIntUntil = allocateSampleVectorIntUntil( 0, initial_state );
	/* Cast to the parent structure which contains initial_state and curr_sample_size */
	PTSampleVec pSampleVecIntUntilBase = (PTSampleVec) pSampleVecIntUntil;
	
	/* Extend the initial vector with sample_size observations*/
	simulateSampleVectorIntUntilCTMC( pStateSpace, pSampleVecIntUntil, sample_size, pNotPhiAndNotPsiBitSet,
					pPhiBitSet, pPsiBitSet, left_time_bound, right_time_bound, pExitRatesOfAllowedStates );
	
	/* printSampleVectorIntUntil( pSampleVecIntUntil ); */
	/* Perform the check of the initial conf. int. */
	computeBordersConfInt( gen_conf_zeta, &leftBorder, &rightBorder, pSampleVecIntUntilBase->curr_sample_size,
				pSampleVecIntUntil->sum_good, pSampleVecIntUntil->sum_good, AGRESTI_COULL_CONF_INT);
	
	/* Check the conf. int. against the probability constraint */
	mc_result = checkBoundVSConfInt( comparator, prob_bound, leftBorder, rightBorder, indiff_width );
	
	/* The main simulation cycle */
	while( ( mc_result == TVL_NN ) && ( sample_size < max_sample_size ) ){
		/* Increase the sample size */
		increment( &sample_size, sample_size_step, max_sample_size );
		
		/* Extend the initial vector with sample_size observations*/
		simulateSampleVectorIntUntilCTMC( pStateSpace, pSampleVecIntUntil, sample_size, pNotPhiAndNotPsiBitSet,
						pPhiBitSet, pPsiBitSet, left_time_bound, right_time_bound, pExitRatesOfAllowedStates );
		
		/* printSampleVectorIntUntil( pSampleVecIntUntil ); */
		/* Perform the check of the initial conf. int. */
		computeBordersConfInt( gen_conf_zeta, &leftBorder, &rightBorder, pSampleVecIntUntilBase->curr_sample_size,
					pSampleVecIntUntil->sum_good, pSampleVecIntUntil->sum_good, AGRESTI_COULL_CONF_INT);
		
		/* Check the conf. int. against the probability constraint */
		mc_result = checkBoundVSConfInt( comparator, prob_bound, leftBorder, rightBorder, indiff_width );
	}
	/* Update the pCiLeftBorders and pCiRightBorders */
	pCiLeftBorders[arr_index] = leftBorder;
	pCiRightBorders[arr_index] = rightBorder;
	
	/* Update the pYesBitSet or pNoBitSet if we have a definite answer */
	markYesNoSetEntree(mc_result, initial_state, pYesBitSet, pNoBitSet);

	*pNumUsedObserv	= pSampleVecIntUntilBase->num_visited_states;
	
	/* Free the sample vectors */
	freeSampleVectorIntUntil( pSampleVecIntUntil );
}

/**
* This function is an implementation of the intervalUntil function from the PhD Thesis
* of Ivan S. Zapreev. The function does the model checking of time-interval until operator
* on the provided CTMC and uses the runtime-simulation settings from the "simulation.h"
* @param pStateSpace the sparse matrix of the embedded DTMC with good and bad
*			states made absorbing
* @param confidence this is is confidence we should use for simulations.
*			Note that since the formulas can be nested one can not use the
*			overall confidence set in simulation.h
* @param pPhiBitSet the bitsets containing all the good absorbing states
* @param pPsiBitSet the bitsets containing all the transient states
* @param left_time_bound the left time bound
* @param right_time_bound the right time bound
* @param comparator the comparator, one of:                                           getSimMaxSampleSize
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param initial_state the initial state index if isSimOneInitState == TRUE
*
* NOTE: Below we have pointers to the pointers, these are going to be the return values
*
* @param ppYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param ppNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param ppCiLeftBorders will contain the left conf. int. borders
* @param ppCiRightBorders will contain the right conf. int. borders
* @param pResultSize the pointer to the size of arrays *ppCiLeftBorders and *ppCiRightBorders
* @param pMaxNumUsedObserv the return-value variable that will store the maximum number
* (over all initial states) of used observations
*/
extern void modelCheckTimeIntervalUntilCTMC( const sparse * pStateSpace, const double * pCTMCRowSums,
						const double confidence, const bitset *pPhiBitSet,
						const bitset * pPsiBitSet, const double left_time_bound,
						const double right_time_bound, bitset ** ppYesBitsetResult,
						bitset ** ppNoBitsetResult, double ** ppProbCILeftBorder,
						double ** ppProbCIRightBorder, int * pResultSize,
						const int comparator, const double prob_bound,
						const int initial_state, const BOOL isSimOneInitState,
						unsigned int * pMaxNumUsedObserv ){
	bitset *pNotPhiAndNotPsiBitSet = NULL, * pTmpBitSet = NULL;
	/* This vector will store the exit rates for the elements of pNotPhiAndNotPsiBitSet*/
	double * pExitRatesOfAllowedStates = NULL;
	/* This array will hold indeces of the possible initial states */
	int *pNAIStatesIdx = NULL;

	/* 1: Compute the set of "Allways illegal" states, i.e. not Psi and not Phi states */
	pNotPhiAndNotPsiBitSet = not( pPhiBitSet );
	pTmpBitSet = not( pPsiBitSet );
	and_result( pTmpBitSet, pNotPhiAndNotPsiBitSet );
	free_bitset( pTmpBitSet );
	
	/* Compute possible initial-state indexes */
	pTmpBitSet = not( pNotPhiAndNotPsiBitSet );
	pNAIStatesIdx = count_set( pTmpBitSet );
	
	/* 2: Obtain the embedded DTMC and the vector of exit rates */
	/* (the diagonal elements stay intact) */
	pExitRatesOfAllowedStates = make_embedded_dtmc_out_of_rate_mtx_vs( pStateSpace, pCTMCRowSums, pNAIStatesIdx );

	/* print_mtx_sparse(pStateSpace); */
	/* printf("The valid-state exit rates are listed here: "); */
	/* print_vec_double( pStateSpace->rows, pExitRatesOfAllowedStates ); */
	/* printf("\n"); */
	
	/* 3: Do model checking via simulations */
	modelCheckStatesCommon( pStateSpace, confidence, ppYesBitsetResult, ppNoBitsetResult,
				ppProbCILeftBorder, ppProbCIRightBorder, pResultSize, comparator,
				prob_bound, initial_state, isSimOneInitState, pTmpBitSet, pMaxNumUsedObserv,
				modelCheckOneStateIUCTMC, pNotPhiAndNotPsiBitSet, pPhiBitSet, pPsiBitSet,
				left_time_bound, right_time_bound, pExitRatesOfAllowedStates );
	
	/* 4: Restore the state space */
	/* NOTE: not required for diagonal elements */
	restore_rate_mtx_out_of_embedded_dtmc_vs( pStateSpace, pNAIStatesIdx, pExitRatesOfAllowedStates );
	
	/* 5: Update the ppYesBitsetResult and ppNoBitsetResult with respect to pNotPhiAndNotPsiBitSet */
	considerAlwaysAndNeverStates( isSimOneInitState, *ppYesBitsetResult, *ppNoBitsetResult,
					NULL, pNotPhiAndNotPsiBitSet, initial_state, comparator, prob_bound );
	
	/* 6: Free the allocated memory */
	free_bitset( pTmpBitSet );
	free_bitset( pNotPhiAndNotPsiBitSet );
	free( pExitRatesOfAllowedStates );
	free( pNAIStatesIdx );
}

/*******************************************************************************/
/********THE FUNCTIONS FOR SIMULATING STEADY-STATE OPERATOR ON A CTMC***********/
/*******************************************************************************/

/**
* This function is designed to simulate the new regeneration cycles for steady
* state CTMC: "S<>p(Psi)".
*
* WARNING: We simulate the states ASSUMING there are no self loops!
*	The absorbing states have to be found based on the absence of transitions to other states
*
* @param pStateSpace the sparse matrix of the embedded DTMC
* @param pSampleVecSteady the sample vector obtained on the previous iteration
* @param old_sample_size the old number of regeneration cycles (sample size)
* @param pPsiBitSet the bitsets containing all Psi states
* @param pExitRatesOfAllowedStates the array of exit rates for the Psi states,
* 	note that its size corresponds to the dimensions of pStateSpace
*/
static void simulateSSSampleCTMC( const sparse * pStateSpace, PTSampleVecSteady pSampleVecSteady,
					const int old_samle_size, const bitset * pPsiBitSet, 
					const double * pExitRatesOfAllowedStates ){
	int i, current_obs_state;
	/* Calculate the expected time spent in the i'th regeneration cycle in here */
	double accumulated_time;
	/* Calculate the accumulated number of "good" states divided by exit rates */
	/* in the i'th regeneration cycle in here */
	double accumulated_good;
	/* Will store the number of newly visited states */
	unsigned int newlyVisitedStates = 0;
	
	/* Cast to the parent structure which contains initial_state and curr_sample_size */
	PTSampleVec pSampleVecSteadyBase = (PTSampleVec) pSampleVecSteady;

	/* For every regeneration cycle we start in the same initial state */
	current_obs_state = pSampleVecSteadyBase->initial_state;

	/* Simulate the newly added regeneration cycles */
	for( i = old_samle_size; i < pSampleVecSteadyBase->curr_sample_size; i++ ){
		/* Set the initial counter values to zero */
		accumulated_time = 0.0;
		accumulated_good = 0.0;
		
		/* Simulate one regeneration cycle, i.e. until the initial state is reached again */
		do{
			accumulated_time += 1.0 / pExitRatesOfAllowedStates[current_obs_state];
			/* Increment only if the current state is a Psi state */
			if( get_bit_val( pPsiBitSet, current_obs_state ) ){ 
				accumulated_good += 1.0 / pExitRatesOfAllowedStates[current_obs_state];
			}
			
			/* Go to the next state */
			current_obs_state = computeNextState( pStateSpace->val[ current_obs_state ],
						pStateSpace->ncols[ current_obs_state ], current_obs_state );

			/* The next state was simulated */
			newlyVisitedStates++;
		} while( current_obs_state != pSampleVecSteadyBase->initial_state );
		
		/* printf("REGENERATION SYCLE IS DONE\n"); */
		
		/* Store the results for this regeneration cycle in the sample structure */
		pSampleVecSteady->pTimeObservationsVec[i] = accumulated_time;
		pSampleVecSteady->pGoodObservationsVec[i] = accumulated_good;
		
	}
	
	/* printSampleVectorSteady( pSampleVecSteady ); */
	
	/* Update the number of simulations used up to this point */
	pSampleVecSteadyBase->num_visited_states += newlyVisitedStates;
}

/**
* This function is designed to extended and simulate the sample for steady state
* CTMC: "S<>p(Psi)". This function is an implementation of the regenerative method
* mentioned in PhD Thesis of Ivan S. Zapreev. It adds new regeneration cycles to 
* the sample and simulates.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC
* @param pSampleVecSteady the sample vector obtained on the previous iteration
* @param new_sample_size the new sample size
* @param pPsiBitSet the bitsets containing all Psi states
* @param pExitRatesOfAllowedStates the array of exit rates for the Psi states,
* 	note that its size corresponds to the dimensions of pStateSpace
*/
static void simulateSampleVectorSSCTMC( const sparse * pStateSpace, PTSampleVecSteady pSampleVecSteady,
					const int new_sample_size, const bitset * pPsiBitSet, 
					const double * pExitRatesOfAllowedStates ){
	IF_SAFETY( pSampleVecSteady != NULL )
		/* Cast to the parent structure which contains initial_state and curr_sample_size */
		PTSampleVec pSampleVecSteadyBase = (PTSampleVec) pSampleVecSteady;
		
		IF_SAFETY( pPsiBitSet != NULL )
			IF_SAFETY( pStateSpace != NULL )
				if( pSampleVecSteadyBase->curr_sample_size < new_sample_size ){
					const int old_sample_size = pSampleVecSteadyBase->curr_sample_size;
					
					/* Add new regeneration cycles */
					extendSampleVectorSteady( pSampleVecSteady, new_sample_size );
					
					/*Do simulations for the newly added observations */
					simulateSSSampleCTMC( pStateSpace, pSampleVecSteady, old_sample_size,
								pPsiBitSet, pExitRatesOfAllowedStates );
				}
			ELSE_SAFETY
				printf("ERROR: The sparse matrix is NULL.\n");
				exit(1);
			ENDIF_SAFETY
		ELSE_SAFETY
			printf("ERROR: The Psi states are undefined.\n");
			exit(1);
		ENDIF_SAFETY
	ELSE_SAFETY
		printf("ERROR: The sample that has to be simulated further is NULL.\n");
		exit(1);
	ENDIF_SAFETY
}


/**
* This function is an implementation of the checkCIHybrid function from
* the PhD Thesis of Ivan S. Zapreev. The function does the recomputation of
* the confidence interval, afterwards check it against the probability 
* constraint p given by the formula "S<>p(Psi)".
* @param pReachableTrivialBSCCs the bitset containing the indices of reachable trivial BSCCs.
* @param pReachableNonTrivBSCCs the bitset containing the indices of reachable non-trivial BSCCs.
* @param pBSCCReachProbability the array containing the reachability probabilities of every BSCC.
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param error_bound the error bounf for the numerically computed reachability
*			probabilities (Hybrid mode).
* @param pLeftCiSSBorder the array containing the left conf. int. border for every BSCC.
* @param pRightCiSSBorder the array containing the right conf. int. border for every BSCC.
* @param indiff_width the width of the indifference region
*
* NOTE: The parameters below will contain the return values

* @param pLeftBorder the overall left conf. int. border for the actual initial state.
* @param pRightBorder the overall right conf. int. border for the actual initial state.
* @return returns one of: TVL_TT, TVL_FF, TVL_NN.
*/
static TV_LOGIC checkConfIntSSHybrid(bitset * pReachableTrivialBSCCs, bitset * pReachableNonTrivBSCCs,
			 	const double * pBSCCReachProbability, const int comparator, const double prob_bound, 
			 	const double error_bound, double * pLeftCiSSBorder, double * pRightCiSSBorder, 
			 	const double indiff_width, double * pLeftBorder, double * pRightBorder){
	int i, curr_index;
	/* The confidence interval borders, which hold the end result finally */
	double LeftCiBorder = 0;
	double RightCiBorder = 0;
	
	int numberOfTrivialBSCCs = pReachableTrivialBSCCs->n;
	
	i = 0, curr_index = -1;
	/* Compute the combined confidence interval for the trivial BSCCs */
	/* First we should iterate through the trivial BSCCs since the reachability probabilities for */
	/* them are located in the pBSCCReachProbability in the very beginning */
	/* this part of the array is "indexed by" the number of BSCCs with Psi states whereas pLeftCiSSBorder */
	/* and pRightCiSSBorder are "indexed by" the number of reachable Psi BSCCs */
	while( ( curr_index =  get_idx_next_non_zero( pReachableTrivialBSCCs, curr_index ) ) != -1 ){
		LeftCiBorder += ( pBSCCReachProbability[curr_index] - error_bound ) * pLeftCiSSBorder[i];
		RightCiBorder += ( pBSCCReachProbability[curr_index] + error_bound ) * pRightCiSSBorder[i];
		i++;	
	}
	
	/* WARNING: At this point we do not change the value of i because we */
	/* continue iterating through the pLeftCiSSBorder and pRightCiSSBorder arrays */
	curr_index = -1;
	/* Compute the combined confidence interval for the non-trivial BSCCs */
	/* Note that to get the indexes of pBSCCReachProbability from pReachableNonTrivBSCCs */
	/* related to the reachable non-trivial Psi BSCCs we have to add numberOfTrivialBSCCs to the index */
	while( ( curr_index =  get_idx_next_non_zero( pReachableNonTrivBSCCs, curr_index ) ) != -1 ){
		LeftCiBorder += ( pBSCCReachProbability[curr_index + numberOfTrivialBSCCs] - error_bound ) * pLeftCiSSBorder[i];
		RightCiBorder += ( pBSCCReachProbability[curr_index + numberOfTrivialBSCCs] + error_bound ) * pRightCiSSBorder[i];
		i++;	
	}
	
	/* Check if we've gotten values < 0.0 and > 1.0 */
	if( LeftCiBorder < 0.0 ){
		LeftCiBorder = 0.0;
	}
	if( RightCiBorder > 1.0 ){
		RightCiBorder = 1.0;
	}
	
	/* Assign the return values for the combined confidence interval borders */
	( * pLeftBorder ) = LeftCiBorder;
	( * pRightBorder ) = RightCiBorder;
	
	/* Check the probability constraint */
	return checkBoundVSConfInt( comparator, prob_bound, LeftCiBorder, RightCiBorder, indiff_width );	
} 

/**
* Model checks the steady-state operator "S<>p(Psi)" for one initial state only.
* Note that this is a univesral procedure that assumes that we work with the embedded DTMC.
* 
* WARNING: We simulate the states ASSUMING there are no self loops! The self loops of absorbing
* states have to be detected by the absence of other outgoing transitions.
* 
* @param pStateSpace the sparse matrix of the embedded DTMC
* @param initial_state the initial state for simulations
* @param indiff_width the width of the indifference region
* @param confidence the desired confidence of the answer
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param arr_index indicates the index in arrays pCiLeftBorders and pCiRightBorders
*			that corresponds to the initial_state
*
* NOTE: The parameters below will contain the return values
*
* @param pCiLeftBorders for storing the left conf. int. border
* @param pCiRightBorders for storing the right conf. int. border
* @param pYesBitSet contains curr_state if it satisfies the formula
* @param pNoBitSet contains curr_state if it does not satisfy the formula
* @param pNumUsedObserv the pointer to the return-value variable that will store the
*		number of observations used for simulating this one state.
* @param pArguments contains the following (ordered!):
*		1. pPsiBitSet the bitset containing the Psi states
* 		2. ppReachProbability The reachability probabilities from all 
			initial states to BSCCs with Psi states.
*		3. error_bound the error bound for the numerically computed reachability
*			probabilities 
*		4. pExitRatesOfAllowedStates the array of exit rates for the
*			"pPsiBitSet" states, note that its size corresponds
*			to the dimensions of pStateSpace
*		5. pSampleVecSteady the array of sample structures for the non-trivial BSCCs
*		6. numberOfNonTrivBSCCs the number of non-trivial BSCCs containing Psi states 
*		7. numberOfTrivialBSCCs the number of trivial BSCCs consisting of good states
*/
static void modelCheckOneStateSSHybridCTMC( const sparse* pStateSpace, const int initial_state, const double indiff_width,
					const double confidence, const int comparator, const double prob_bound,
					const int arr_index, double *pCiLeftBorders, double *pCiRightBorders,
					bitset * pYesBitSet, bitset * pNoBitSet, unsigned int *pNumUsedObserv,
					va_list pArguments ){
	/* Contains the pPsiBitSet */
	const bitset * pPsiBitSet = va_arg( pArguments, bitset * );
	/* The reachability probabilities from all initial states to BSCCs with Psi states */
	/* e.g. ppReachProbability[0][0] contains the probability to reach BSCC 0 from state 1 */
	/* ppReachProbability[0][1] contains the probability to reach BSCC 0 from state 2 etc. */
	double ** ppReachProbability = va_arg( pArguments, double ** );
	/* Contains the error bound for numerical computation */
	const double error_bound = va_arg( pArguments, double );
	/* Contains the exit rates for the "pPsiBitSet" states */
	const double * pExitRatesOfAllowedStates = va_arg( pArguments, double * );
	/* Contains a sample structure for every non-trivial BSCC */
	PTSampleVecSteady * pSampleVecSteady = va_arg( pArguments, PTSampleVecSteady * );
	/* The number of non-trivial BSCCs containing Psi states */
	const int numberOfNonTrivBSCCs = va_arg( pArguments, int );
	/* The number of trivial BSCCs containing Psi states */
	const int numberOfTrivialBSCCs = va_arg( pArguments, int );
	/* The average number of states in non-trivial BSCCs */
	const int numberOfNonTrivBSCCStates = va_arg( pArguments, int );
	
	/* The number of BSCCs containing Psi states */
	int numberOfBSCCs = numberOfNonTrivBSCCs + numberOfTrivialBSCCs;
	/* The model checking result*/
	TV_LOGIC mc_result = TVL_NN;
	/* The combined conf. int. left and right borders */
	double leftBorder = 0.0, rightBorder = 1.0;
	int i, curr_index;
	/* The bitset indicating the trivial BSCCs containing Psi states, which are also reachable */
	bitset * pReachableTrivialBSCCBitSet;
	/* The bitset indicating the non-trivial BSCCs containing Psi states, which are also reachable */
	bitset * pReachableNonTrivBSCCBitSet;
	/* The number of non-trivial BSCCs, that contain Psi states and are reachable from the */
	/* initial state */
	int numberOfReachableNonTrivBSCCs = 0;
	/* The number of trivial BSCCs, that contain Psi states and are reachable from the */
	/* initial state */
	int numberOfReachableTrivialBSCCs = 0;
	/* The number of ovreall BSCCs containing Psi states, that are reachable from the initial state */
	int numberOfReachableBSCCs = 0;
	
	/* The array holding the reachability probabilities from the initial state */
	/* to the reachable BSCCs. The first numberOfTrivialBSCCs are devoted to the */
	/* reachability probs. of trivial BSCCs and the rest numberOfNonTrivBSCCs */
	/* elements for the reachability probs. of non-trivial BSCCs*/
	double *pBSCCReachProbability = (double *) calloc( numberOfBSCCs , sizeof(double) );
	/* The pBSCCReachProbability will store all the probabilities of reaching */
	/* all the BSCCs from the given initial state. Some of them are going to be zero */
	/* The indexes of non zero elements of pBSCCReachProbability are going to be stored */
	/* in the next two bitsets. Note that they are indexed independently, i.e. the indexes */
	/* of both bitsets will start from zero, and to get the pBSCCReachProbability index from */
	/* pReachableNonTrivBSCCBitSet we will need to add numberOfTrivialBSCCs to it */
	pReachableTrivialBSCCBitSet = get_new_bitset( numberOfTrivialBSCCs );
	pReachableNonTrivBSCCBitSet = get_new_bitset( numberOfNonTrivBSCCs );
	
	/* Having reachability probabilities in place, we know what BSCCs, containing Psi states, */
	/* are reachable with non-zero probability. */
	/* WARNING: There might be problems with precision, i.e. 0.0 probability might be represented */
	/* as a very vey small, but non-zero value, like 0.000000000000000000000000000234234234 this */
	/* will have to be checked, at present we assume that it JUST works. */
	
	/* The BSCC reachability probabilities for every BSCC i*/
	/* For the trivial BSCCs found */
	for( i = 0; i < numberOfTrivialBSCCs; i++ ){
		/* If BSCC i is reachable from the initial state, increment reachable BSCC counter */
		if( ( pBSCCReachProbability[i] = ppReachProbability[i][initial_state] ) > 0.0 ){
			numberOfReachableTrivialBSCCs++;
			set_bit_val( pReachableTrivialBSCCBitSet, i, BIT_ON );
		}
	}
	
	/* For the non-trivial BSCCs found */
	for( i = numberOfTrivialBSCCs; i < numberOfBSCCs; i++ ){
		/* If non-trivial BSCC i is reachable from the initial state, increment reachable non-trivial */
		/* BSCC counter */
		if( ( pBSCCReachProbability[i] = ppReachProbability[i][initial_state] ) > 0.0 ){
			numberOfReachableNonTrivBSCCs++;
			set_bit_val(pReachableNonTrivBSCCBitSet, i - numberOfTrivialBSCCs, BIT_ON);
		}
	}

	numberOfReachableBSCCs = numberOfReachableNonTrivBSCCs + numberOfReachableTrivialBSCCs;
		
	/* Do computations only if there is at least one reachable BSCC, that contains Psi states */
	if ( numberOfReachableBSCCs > 0 ){
		/* Compute the confidence of this sub-simulation as the n'th root of the overall */
		/* confidence */
		const double sub_confidence = ROOT(confidence, numberOfReachableNonTrivBSCCs); 
		/* Compute the general confidence zeta value */
		const double gen_conf_zeta = gsl_cdf_ugaussian_Pinv( sqrt( sub_confidence ) );
		/* NOTE: The confidence we have is used to derive z for the left and the */
		/* right border BUT the TRUE/FALSE answer is based only on one border of */
		/* the conf. int., i.e. we could use a smaller confidence for model checking, */
		/* but then again we have a different confidence for the produced conf. int. */
		/* this is bad because then the user does not know that. */
		
		/* The conf. int. left and right borders for the simulation part */
		double *pLeftBorder = (double *) calloc( numberOfReachableBSCCs, sizeof(double) );
		double *pRightBorder = (double *) calloc( numberOfReachableBSCCs, sizeof(double) );
		
		/* Initialize the sample size parameters */
		int sample_size = getSimMinSampleSize();
		int sample_size_step;
		const int max_sample_size = getSimMaxSampleSize();
		
		/* Is the sample size step set manually or automatically? */
		if( isSimSampleSizeStepAuto() ){
			/* For automatic sample size step calculate the initial sample size step */
			sample_size_step = sqrt( numberOfNonTrivBSCCStates );
		} else {
			/* For manual sample size step get the value set */
			sample_size_step = getSimSampleSizeStep();
		}		
			
		/* ...and their initialization, First for the trivial BSCCs */
		for( i = 0; i < numberOfReachableTrivialBSCCs; i++ ){
			/* If the BSCC contains only one state, the long run probability is exactly 1 */
			pLeftBorder[i] = pRightBorder[i] = 1.0;
		}
		/* Next for the non-trivial BSCCs*/
		/* NOTE: There is no need to do pLeftBorder[i] = 0.0 because we use calloc */
		for(i = numberOfReachableTrivialBSCCs; i < numberOfReachableBSCCs; i++){
			pRightBorder[i] = 1.0;
		}
		
		/* Is a definite answer already providable without simulating? */
		mc_result = checkConfIntSSHybrid( pReachableTrivialBSCCBitSet, pReachableNonTrivBSCCBitSet,
						pBSCCReachProbability, comparator, prob_bound, error_bound, 
						pLeftBorder, pRightBorder, indiff_width, &leftBorder, &rightBorder);
		
		if( ( mc_result == TVL_NN ) && ( numberOfReachableBSCCs - numberOfReachableTrivialBSCCs > 0 ) ){
			/* Perform the main simulation cycle */
			do{
				/* The average length of cycles simulated in this step */
				int avg_cycle_length = 0;
				
				/* Increment the sample size, i.e. the number of regeneration cycles */
				increment( &sample_size, sample_size_step, max_sample_size );
			
				/* Simulate for every BSCC its sample vector and compute the new confidence borders*/
				curr_index = -1; i = numberOfReachableTrivialBSCCs;
				while( ( curr_index = get_idx_next_non_zero( pReachableNonTrivBSCCBitSet, curr_index ) ) != -1 ){
					simulateSampleVectorSSCTMC(pStateSpace, pSampleVecSteady[curr_index], sample_size,
									pPsiBitSet, pExitRatesOfAllowedStates );
					computeBordersConfIntSS( sample_size, gen_conf_zeta, &pLeftBorder[i],
								&pRightBorder[i], pSampleVecSteady[curr_index]);
					i++;
					avg_cycle_length += ((PTSampleVec) pSampleVecSteady[curr_index])->num_visited_states ;
				}
				
				/* Calculate the average cycle length per BSCC and completed cycle */
				avg_cycle_length = avg_cycle_length / ( ( i - numberOfReachableTrivialBSCCs ) * sample_size );
				/* Is the sample size step set manually or automatically? */
				if( isSimSampleSizeStepAuto() ){
					/* For automatic sample size step calculate the new sample size step */
					sample_size_step = ( numberOfNonTrivBSCCStates / avg_cycle_length ) + sample_size;
				}
				
				/* Compute the combined confidence borders of the numerical and simulation part */
				mc_result = checkConfIntSSHybrid( pReachableTrivialBSCCBitSet, pReachableNonTrivBSCCBitSet,
						pBSCCReachProbability, comparator, prob_bound, error_bound, 
						pLeftBorder, pRightBorder, indiff_width, &leftBorder, &rightBorder);
			}while( ( mc_result == TVL_NN ) && ( sample_size < max_sample_size ) );
		}

		/* Report the number of states visited while this simulation run */
		/* Note that, since the samples are reused, we take into account all the states that */
		/* had to be visited in order to give the answer to this particular initial state */
		curr_index = -1; *pNumUsedObserv = 0;
		while( ( curr_index = get_idx_next_non_zero( pReachableNonTrivBSCCBitSet, curr_index ) ) != -1 ){
			*pNumUsedObserv += ( (PTSampleVec) pSampleVecSteady[curr_index])->num_visited_states;
		}
		
		/* Free memory */
		free(pLeftBorder);
		free(pRightBorder);
	} else {
		/* There is no reachable BSCC containing Psi states, therefore no simulation is needed. */
		/* The long run probability is zero! */
		leftBorder = rightBorder = 0.0;
		mc_result = ( doesComparisonHold( rightBorder, comparator, prob_bound) ) ? TVL_TT : TVL_FF ;
	}

	/* Update the pCiLeftBorders and pCiRightBorders */
	pCiLeftBorders[arr_index] = leftBorder;
	pCiRightBorders[arr_index] = rightBorder;
	
	/* Update the pYesBitSet or pNoBitSet if we have a definite answer */
	markYesNoSetEntree(mc_result, initial_state, pYesBitSet, pNoBitSet);
	
	/*Free memory */
	free( pBSCCReachProbability );
	free_bitset( pReachableTrivialBSCCBitSet );
	free_bitset( pReachableNonTrivBSCCBitSet );
}

/**
* This function is an implementation of the steadyStateHybrid function from
* the PhD Thesis of Ivan S. Zapreev. The function does the model
* checking of the steady-state operator on the provided CTMC and uses the
* runtime-simulation settings from the "simulation.h"
* @param pStateSpace the sparse matrix of the CTMC
* @param confidence this is the confidence we should use for simulations.
* 
* NOTE: that since the formulas can be nested one can not use the
* overall confidence set in simulation.h
* 
* @param pPsiBitSet the bitsets containing the Psi states for S<>p(Psi)
* @param comparator the comparator, one of:
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param initial_state the initial state index if isSimOneInitState == TRUE
* @param pFNumUnbUntilCTMC the function for computing the unbounded-until
*				probabilities numerically (Hybrid mode).
* @param pFAllowedBSCCSearchCTMC the function that is used for searching the BSCCs that contain Psi states.
* @param error_bound the error bounf for the numerically computed reachability
*			probabilities (Hybrid mode).
*
* NOTE: Below we have pointers to the pointers, these are going to be the return values
*
* @param ppYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param ppNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param ppCiLeftBorders will contain the left conf. int. borders
* @param ppCiRightBorders will contain the right conf. int. borders
* @param pResultSize the pointer to the size of arrays *ppCiLeftBorders and *ppCiRightBorders
* @param pMaxNumUsedObserv the return-value variable that will store the maximum number
* (over all initial states) of used observations
*/
void modelCheckSteadyStateHybridCTMC( sparse * pStateSpace, const double * pCTMCRowSums ,
					const double confidence, bitset * pPsiBitSet,
					bitset ** ppYesBitsetResult, bitset ** ppNoBitsetResult,
					double ** ppProbCILeftBorder, double ** ppProbCIRightBorder,
					int * pResultSize, const int comparator, const double prob_bound,
					const int initial_state, const BOOL isSimOneInitState,
					const TPFunctNumUnbUntilCTMC pFNumUnbUntilCTMC,
					const TPFunctAllowedBSCCSearchCTMC pFAllowedBSCCSearchCTMC,
					const double error_bound, unsigned int * pMaxNumUsedObserv){
	int i, j;
	int index;
	int initial_state_sim;
	/* The multiplier for the choice of the sample-based regeneration point */
	int sampleBSCCDimensionMultiplier;
	bitset * pTmpBitSet = NULL;
	/* The number of states in the BSCC currently handled */
	int numberOfBSCCStates;
	/* The number of BSCCs containing Psi states */
	int numberOfBSCCs;
	/* The number of non-trivial BSCCs containing Psi states */
	int numberOfNonTrivBSCCs = 0;
	/* The average number of states in non-trivial BSCCs containing Psi states */
	int numberOfNonTrivBSCCStates = 0;
	/* The number of trivial BSCCs containing Psi states */
	int numberOfTrivialBSCCs = 0;
	/* The structure, that stores the set of states belonging to non-trivial BSCC with array index i */
	bitset ** ppNonTrivBSCCBitSets = NULL;
	/* The structure, that stores the set of states belonging to non-trivial BSCCs */
	bitset * pNonTrivBSCCStatesBitSet = NULL;
	/* The bitset indicating the Psi states belonging to trivial BSCCs */
	bitset * pTrivialBSCCBitSet = NULL;
	/* The array that holds the indeces of states belonging to non-trivial BSCCs */
	int * pNonTrivBSCCStates = NULL;
	/* The array that holds the number state occurrences while simulating to fix the regeneration state */
	int * pSimStateOccurrence;
	/* The structure, that stores the probability to reach BSCC i from state j */
	double ** ppReachProbability;
	/* This vector will store the exit rates for the elements of pPsiBitSet*/
	double * pExitRatesOfAllowedStates = NULL;
	/* The state space dimension */
	const int n_states = pStateSpace->rows;
	
	/* The structure, which stores the samples */
	PTSampleVecSteady *pSampleVecSteady = NULL;
	
	/* Create True bitset*/
	bitset *pTRUEBitSet = get_new_bitset( n_states );
	/* The bitset that holds the possible initial states */
	bitset *pValidInitialStatesBitSet = get_new_bitset( n_states );
	/* All states are possible initial states */
	fill_bitset_one(pValidInitialStatesBitSet);
	/* Fill True bitset with ones */
	fill_bitset_one( pTRUEBitSet );
	/* Create bitset, that holds all states belonging to non-trivial BSCCs in the end */
	pNonTrivBSCCStatesBitSet = get_new_bitset( n_states );
	
	/* Find BSCCs that contain Psi states, divided into the set of trivial */ 
	/* and non-trivial BSCCs */
	pTrivialBSCCBitSet = pFAllowedBSCCSearchCTMC( pStateSpace, pPsiBitSet, &ppNonTrivBSCCBitSets, &numberOfNonTrivBSCCs );
	numberOfTrivialBSCCs = count_non_zero( pTrivialBSCCBitSet );
	/* Compute the total number of BSCCs that contain Psi states */
	numberOfBSCCs = numberOfNonTrivBSCCs + numberOfTrivialBSCCs;
	
	/* For the non-trivial BSCCs containing Psi states */
	for( i = 0; i < numberOfNonTrivBSCCs; i++ ){
		/* Create a bitset that holds all states belonging to non-trivial BSCCs */
		or_result( ppNonTrivBSCCBitSets[i], pNonTrivBSCCStatesBitSet );
	}
	/* Fill the array with the indeces of states belonging to non-trivial BSCCs */
	pNonTrivBSCCStates = count_set( pNonTrivBSCCStatesBitSet );
	
	/* Obtain the embedded DTMC and the vector of exit rates */
	/* WARNING: The self loops on states are not modified. Thus the proper way */
	/* to identify an absorbing state is to make sure that there are */
	/* no non-diagonal elements in the corresponding row. */
	pExitRatesOfAllowedStates = make_embedded_dtmc_out_of_rate_mtx_vs( pStateSpace, pCTMCRowSums, pNonTrivBSCCStates );
	
	/* The structure, which stores the samples */
	/* NOTE: We only create a sample vector for non-trivial BSCCs */
	pSampleVecSteady = (PTSampleVecSteady *) calloc( numberOfNonTrivBSCCs, sizeof( PTSampleVecSteady ) );
	sampleBSCCDimensionMultiplier = getSampleBSCCDimensionMultiplier();
	/* Allocate an empty sample vector for every reachable non-trivial BSCC containing Psi states */
	/* This array of sample vectors contains the samples corresponding to the non-trivial BSCCs */
	/* The order is important, i.e. it is the same order of BSCCs as in ppNonTrivBSCCBitSets */
	for( i = 0; i < numberOfNonTrivBSCCs; i++){
			
		numberOfBSCCStates = count_non_zero( ppNonTrivBSCCBitSets[i] );
		/* Count the overall number of states in non-trivial BSCCs */
		numberOfNonTrivBSCCStates += numberOfBSCCStates;
			
		/* If pure regenerative method is used, the initial state for the regenerative method */
		/* is choosen "randomly", here: the BSCC state with the smallest state number (index) */
		if( isRegMethodModePure() ){
			initial_state_sim = get_idx_next_non_zero( ppNonTrivBSCCBitSets[i], -1 );
		
			/* Allocate & initialize the structure that holds the sample vectors */
			pSampleVecSteady[i]  = allocateSampleVectorSteady( 0, initial_state_sim );
		/* If heuristic regenerative method is used, the initial state is choosen according to */
		/* the heuristic: simulate a sample for every BSCC with length */
		/* sampleBSCCDimensionMultiplier x number of states in the BSCC, taking the most visited */
		/* state within this sample as regeneration state */ 
		} else if( isRegMethodModeHeuristic() ){
			initial_state_sim = get_idx_next_non_zero( ppNonTrivBSCCBitSets[i], -1 );
			pSimStateOccurrence = (int *) calloc( n_states, sizeof( int ) );
			pSimStateOccurrence[initial_state_sim]++;
		
			/* Simulate a sample of length sampleBSCCDimensionMultiplier x number of states in the BSCC */
			/* Store the number of occurrences for every state */
			for(j = 0; j < numberOfBSCCStates * sampleBSCCDimensionMultiplier; j++){
				initial_state_sim = computeNextState( pStateSpace->val[ initial_state_sim ],
						pStateSpace->ncols[ initial_state_sim ], initial_state_sim );
				pSimStateOccurrence[ initial_state_sim ]++;
			}
			/* Choose the state as regeneration state, that was most visited during the sample */
			for(j = 0; j < n_states; j++){
				if(pSimStateOccurrence[j] > pSimStateOccurrence[initial_state_sim]){
					initial_state_sim = j;
				}
			}
			/* Allocate & initialize the structure that holds the sample vectors */
			pSampleVecSteady[i]  = allocateSampleVectorSteady( 0, initial_state_sim );
		
			/* Free memory */
			free(pSimStateOccurrence);
		}
	}
	/* Calculate the average number of states in a BSCC */
	if( numberOfNonTrivBSCCs > 0 ){
		numberOfNonTrivBSCCStates = numberOfNonTrivBSCCStates / numberOfNonTrivBSCCs;
	}
	
	/* Compute the reachability probabilities for these BSCCs*/
	/* In this step, due to global model checking we will have*/
	/* reachability probabilities for all initial states.*/
	
	/* Initialize the structure, that stores the probability to reach BSCC i from state j */
	/* The first numberOfTrivialBSCCs elements will contain pointers to arrays of */
	/* reachability probabilities for the trivial BSCCs, in the same order thay are */
	/* located in the pTrivialBSCCBitSet array then the probabilities for BSCCs */
	/* stored in the ppNonTrivBSCCBitSets array */
	ppReachProbability = (double **) calloc( numberOfBSCCs, sizeof( double * ) );

	/* Compute the probabilities P(true U BSCC) */
	index = -1; i = 0;
	
	/* For the trivial BSCCs containing Psi states */
	pTmpBitSet = get_new_bitset( n_states );
	while( ( index = get_idx_next_non_zero( pTrivialBSCCBitSet, index ) ) != -1 ){
		set_bit_val( pTmpBitSet, index, BIT_ON );
		
		/* WARNING: We have to be careful here because the state space is an implicit parameter */
		/* We use the state space accessible from the get_state_space() function of runtime.h */
		ppReachProbability[i] = pFNumUnbUntilCTMC( pTRUEBitSet, pTmpBitSet);

		set_bit_val( pTmpBitSet, index, BIT_OFF );
		i++;
	}
	free_bitset(pTmpBitSet); pTmpBitSet = NULL;

	/* For the non-trivial BSCCs containing Psi states */
	for( i = numberOfTrivialBSCCs; i < numberOfBSCCs; i++ ){
		ppReachProbability[i] = pFNumUnbUntilCTMC( pTRUEBitSet, ppNonTrivBSCCBitSets[i - numberOfTrivialBSCCs] );
	}
	
	modelCheckStatesCommon( pStateSpace, confidence, ppYesBitsetResult, ppNoBitsetResult, ppProbCILeftBorder,
				ppProbCIRightBorder, pResultSize, comparator, prob_bound, initial_state, 
				isSimOneInitState, pValidInitialStatesBitSet, pMaxNumUsedObserv,
				modelCheckOneStateSSHybridCTMC, pPsiBitSet, ppReachProbability, error_bound,
				pExitRatesOfAllowedStates, pSampleVecSteady, numberOfNonTrivBSCCs, numberOfTrivialBSCCs, numberOfNonTrivBSCCStates );
		
	/* Restore the state space */
	/* NOTE: The diagonal elements do not have to be restored since thay were not changed */
	restore_rate_mtx_out_of_embedded_dtmc_vs( pStateSpace, pNonTrivBSCCStates, pExitRatesOfAllowedStates );
	free( pExitRatesOfAllowedStates ); pExitRatesOfAllowedStates = NULL;
	free( pNonTrivBSCCStates ); pNonTrivBSCCStates = NULL; 
	free_bitset( pTrivialBSCCBitSet ); pTrivialBSCCBitSet = NULL;
	free_bitset( pValidInitialStatesBitSet ); pValidInitialStatesBitSet = NULL;
	free_bitset( pTRUEBitSet ); pTRUEBitSet = NULL;
	free_bitset( pNonTrivBSCCStatesBitSet ); pNonTrivBSCCStatesBitSet = NULL;
	
	for( i = 0; i < numberOfNonTrivBSCCs; i++ ){
		free_bitset( ppNonTrivBSCCBitSets[i] );
	}
	free( ppNonTrivBSCCBitSets ); ppNonTrivBSCCBitSets = NULL;

	for( i = 0; i < numberOfBSCCs; i++ ){
		free( ppReachProbability[i] );
	}
	free( ppReachProbability ); ppReachProbability = NULL;

	for( i = 0; i < numberOfNonTrivBSCCs; i++){
		freeSampleVectorSteady( pSampleVecSteady[i] );
	}
	free( pSampleVecSteady ); pSampleVecSteady = NULL;
}

/**
* This function is an implementation of the steadyState function from
* the PhD Thesis of Ivan S. Zapreev. The functions do the model
* checking of unbounded-until operator on the provided CTMC and uses the
* runtime-simulation settings from the "simulation.h"
* @param pStateSpace the sparse matrix of the embedded DTMC with good and bad
*			states made absorbing
* @param confidence this is is confidence we should use for simulations.
*			Note that since the formulas can be nested one can not use the
*			overall confidence set in simulation.h
* @param pPsiBitSet the bitsets containing the Psi states for S<>p(Psi)
* @param comparator the comparator, one of:
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param initial_state the initial state index if isSimOneInitState == TRUE
*
* NOTE: Below we have pointers to the pointers, these are going to be the return values
*
* @param ppYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param ppNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param ppCiLeftBorders will contain the left conf. int. borders
* @param ppCiRightBorders will contain the right conf. int. borders
* @param pResultSize the pointer to the size of arrays *ppCiLeftBorders and *ppCiRightBorders
* @param pMaxNumUsedObserv the return-value variable that will store the maximum number
* (over all initial states) of used observations
*/
extern void modelCheckSteadyStatePureCTMC( const sparse * UNUSED(pStateSpace), const double * UNUSED(pCTMCRowSums) ,
						const double UNUSED(confidence), const bitset * UNUSED(pPsiBitSet),
						bitset ** UNUSED(ppYesBitsetResult), bitset ** UNUSED(ppNoBitsetResult),
						double ** UNUSED(ppProbCILeftBorder), double ** UNUSED(ppProbCIRightBorder),
						int * UNUSED(pResultSize), const int UNUSED(comparator), const double UNUSED(prob_bound),
						const int UNUSED(initial_state), const BOOL UNUSED(isSimOneInitState),
						unsigned int * UNUSED(pMaxNumUsedObserv) ){
	/*TODO: Implement*/
}
